from santoku.sql.mysql import MySQLConnector


"""
TODO: investigate the possibility of mocking MySQL for testing
"""


class TestMySQLConnector:
    def test_get_query_results(self):
        pass

    def test_read_sql_query_to_pandas_dataframe(self):
        pass
