from typing import List

from santoku.google.bigquery import BigQueryHandler

"""
TODO: investigate the possibility of mocking BigQuery (and other Google services) for testing
"""


class TestBigqueryHandler:
    def test_run_query(self):
        pass
